from .BaseNN import nn,pth_info
import torch
# import torch.nn as nn
from torch.autograd import Variable
from .version import __version__ 
__all__    = ['nn','torch',  'Variable',' __version__','pth_info'] 